package cz.matejkripner;

import cz.matejkripner.services.MeasurementException;
import cz.matejkripner.services.Result;
import cz.matejkripner.services.Service;
import cz.matejkripner.services.download.DownloadService;
import cz.matejkripner.services.ping.PingService;
import cz.matejkripner.user.ConnectionQuality;
import cz.matejkripner.user.Values;

import java.time.LocalTime;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Class providing simple manipulating with all services together. If you provide {@link Values}, it'll fill it with
 * real values.
 *
 * @author Matej Kripner <kripnermatej@gmail.com>
 * @version 1.0
 */
public class Measurement {

    /**
     * Number of measurements to get average result.
     */
    private static final int MEASUREMENTS = 3;
    /**
     * If measurement got wrong, we can replace it. This is the max enabled number of doing it.
     */
    private static final int MAX_MEASUREMENTS_ADD = 2;

    // Services to manage
    private final PingService pingService;
    private final DownloadService downloadService;
//    private final UploadService uploadService;

    /**
     * Values to be filled in.
     */
    private final Values values;

    /**
     * Executor service for executing measurement parallel.
     */
    private ExecutorService executorService;

    public Measurement(Values values) {
        this.values = values;
        pingService = new PingService();
        downloadService = new DownloadService();
//        uploadService = new UploadService();

    }

    /**
     * Do the measurements and fill the values given in constructor with measured average values.
     */
    public void measure() {
        executorService = Executors.newFixedThreadPool(3);
        values.forEach(property -> property.setRunning(true));
        measureAndSet(pingService, values.latency(), MEASUREMENTS);
        measureAndSet(downloadService, values.downloadSpeed(), MEASUREMENTS);
//        measureAndSet(uploadService, values.downloadSpeed(), MEASUREMENTS);
        executorService.shutdown();
    }


    /**
     * Helper method to do measurements for only one service and set measured values to given property.
     *
     * @param service      to be measured with
     * @param property     part of {@link Values} to be filled in with measured average values
     * @param measurements number of measurements to be done
     * @see cz.matejkripner.user.Values.Property
     */
    private void measureAndSet(Service service, Values.Property property, int measurements) {
        Runnable runnable = () -> {
            try {
                Result result = measure(service, property, measurements); // do the stuff
                property.setInformation(result.toHumanText()); // set information string
                property.setRunning(false);
            } catch (MeasurementException e) {
                if (e.isFatal()) fatalError(e); // other measurements cannot be done
                else localError(property);
            }
        };
        executorService.execute(runnable);
    }

    /**
     * This should be called if an error occurred and other measurements cannot be done.
     *
     * @param e the error
     */
    private void fatalError(Exception e) {
        values.setWarning(e.getMessage()); // information for the user
        values.forEach(property -> {
            property.setRunning(false); // stop all services
            property.setInformation(ConnectionQuality.NO_CONNECTION);
        });
    }

    /**
     * This should be called if an error occurred but other measurements can be done.
     *
     * @param property the error
     */
    private void localError(Values.Property property) {
        property.setRunning(false);
        property.setInformation("An error occur!", cz.matejkripner.user.Color.ERROR.get());
    }

    /**
     * Do the measurements. The given property is there only to set progress. Any information has to be got from
     * returned {@link Result}.
     *
     * @param service      to be measured with
     * @param property     part of {@link Values} to be inform about progress
     * @param measurements number of measurements to be done
     * @return result with measured values
     */
    public Result measure(Service service, Values.Property property, int measurements) {
        Result[] results = new Result[measurements];
        int badMeasurements = 0; // number of uncompleted measurements

        for (int i = 1; i <= results.length; i++) {
            results[i - 1] = service.measure();
            if (!results[i - 1].everythingOK()) { // uncompleted
                if (badMeasurements++ > MAX_MEASUREMENTS_ADD)
                    throw new MeasurementException("Your connection is too bad to be measured!");
                else i--; // in tolerant, try it again
            }
            property.setProgress((double) i / measurements);
        }
        return results[0].merge(results);
    }

    /**
     * If there is a reason to be warned, return this warning, null otherwise.
     *
     * @return warning, if there is a reason to be warned; null otherwise
     */
    public static String warning() {
        if (!System.getProperty("os.name").toLowerCase().contains("win"))
            return "This application is designed for Windows and would not work properly on other OS.";
        if (LocalTime.now().getHour() > 23 || LocalTime.now().getHour() < 4)
            return "It's late, go to the bed instead of testing this.";
        return null; // everything ok
    }

}
