package cz.matejkripner.services;

import cz.matejkripner.Util;

import java.io.IOException;
import java.io.InputStream;
import java.time.LocalTime;
import java.time.temporal.ChronoUnit;
import java.util.concurrent.TimeUnit;

/**
 * Used for reading output of a process.
 *
 * @author Matej Kripner <kripnermatej@gmail.com>
 * @version 1.0
 */
public class ProcessReader {

    /**
     * If a timeout is not specified, this is used.
     */
    private static final long DEFAULT_TIMEOUT = TimeUnit.SECONDS.toMillis(20); // 20 seconds

    /**
     * Simply factory method.
     *
     * @return new ProcessReader
     */
    public static ProcessReader getReader() {
        return new ProcessReader();
    }

    /**
     * Read the output from given process with default timeout.
     *
     * @param process process to be observed
     * @return output of given process
     */
    public String read(Process process) {
        return read(process, DEFAULT_TIMEOUT);
    }

    /**
     * Read the output from given process with given timeout.
     *
     * @param process process to be observed
     * @param timeout timeout of the operation
     * @return output of given process
     */
    public String read(Process process, long timeout) {
        if (timeout <= 0) throw new IllegalArgumentException("Timeout has to be greater than 0");

        InputStream inputStream = process.getInputStream(); // to get the output
        StringBuilder result = new StringBuilder(); // to build the result
        LocalTime now = LocalTime.now(); // to measure the time
        final boolean[] ok = {false}; // if the reading was done successful (array because of lambda)
        Thread reader = new Thread(() -> { // thread for reading the output
            while (now.until(LocalTime.now(), ChronoUnit.MILLIS) < timeout) {
                try {
                    int available;
                    while ((available = inputStream.available()) > 0) { // read available bytes
                        byte[] in = new byte[available];
                        inputStream.read(in);
                        result.append(new String(in));
                    }
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
                if (!process.isAlive()) break; // all output is read and the process is dead
            }
            ok[0] = true; // done successful
        });

        reader.start();
        Util.join(reader); // join the reader
        if (process.isAlive()) { // reader ended but the process not -> error somewhere
            process.destroy();
            throw new ReadingException("Something is terribly wrong, reading result last more than " + timeout + "ms");
        }
        if (!ok[0])
            throw new ReadingException("Something got terribly wrong with result reading!"); // something got wrong
        return result.toString();
    }

    /**
     * Should be thrown when something with reading got wrong.
     */
    public class ReadingException extends RuntimeException {
        public ReadingException(String msg) {
            super(msg);
        }
    }
}
