package cz.matejkripner.services;

import java.util.function.Function;

/**
 * Base class for all result-representing-classes in the application. If fact this is a container for measured values
 * with some additional functions.
 *
 * @author Matej Kripner <kripnermatej@gmail.com>
 * @version 1.0
 */
public abstract class Result {

    /**
     * Merge given result and return the <strong>average</strong> result.
     *
     * @param results to be merged
     * @return the average result
     * @throws NotCompatibleException if the results cannot be merged
     */
    public abstract Result merge(Result... results) throws NotCompatibleException;

    /**
     * Returns true if and only if this result is OK (well-measured), false otherwise.
     *
     * @return true if and only if this result is OK (well-measured); false otherwise
     */
    public abstract boolean everythingOK();

    /**
     * Converts this result to human-readable text.
     *
     * @return to human-readable text
     * @throws MeasurementException if we find out that this cannot be formatted
     */
    public abstract String toHumanText() throws MeasurementException;

    /**
     * If the given result cannot be merged, throws the appropriate exception.
     *
     * @param result to be checked
     * @throws NotCompatibleException if the given result cannot be merged
     */
    protected void checkCompatible(Result result) throws NotCompatibleException {
        if (!(getClass().equals(result.getClass())))
            throw new Result.NotCompatibleException(Result.NotCompatibleException.ANOTHER_TYPE);
        if (!result.everythingOK()) throw new Result.NotCompatibleException(Result.NotCompatibleException.WERE_NOT_OK);
    }

    /**
     * Help method for merging.
     *
     * @param getUnit   get the unit to be averaged from a result
     * @param getResult construct a result from averaged unit
     * @param results   to be merged
     * @param <T>       type of results
     * @return average result
     * @throws NotCompatibleException if the results cannot be merged
     */
    protected <T extends Result> T mergeHelp(Function<T, Integer> getUnit, Function<Integer, T> getResult, Result... results) throws NotCompatibleException {
        int sum = 0;
        for (Result result : results) {
            if (result == null) continue;
            checkCompatible(result);
            sum += getUnit.apply((T) result);
        }
        return getResult.apply(sum / results.length);
    }

    /**
     * Should be thrown when results cannot be merged. Also provide basics messages for the user.
     */

    public class NotCompatibleException extends RuntimeException {

        public static final String WERE_NOT_OK = "Results couldn't be merged, cause one or more of them was not OK";

        public static final String ANOTHER_TYPE = "Results couldn't be merged, cause they weren't of same type";

        public NotCompatibleException(String msg) {
            super(msg);
        }

    }
}
