package cz.matejkripner.services.ping;

import cz.matejkripner.services.Result;
import cz.matejkripner.user.ConnectionQuality;

/**
 * Result of a {@link PingService} measuring. Mainly contains the network latency.
 *
 * @author Matej Kripner <kripnermatej@gmail.com>
 * @version 1.0
 */
public class PingResult extends Result {

    private final String src;
    private final int avgLatency; // in ms
    private final boolean ok;

    public PingResult(String src, int avgLatency, boolean ok) { // if src is null, this is result from merging -> there is no src
        this.src = src;
        this.avgLatency = avgLatency;
        this.ok = avgLatency != 0 && ok;
    }

    public PingResult(String src) {
        this.src = src;
        avgLatency = avgSpeed();
        ok = avgLatency != 0 && ok();
    }

    /**
     * Determine if pinging was done correctly from given ping output.
     *
     * @return pinging was done correctly
     */
    private boolean ok() {
        return src.contains("0%"); // 0% loss
    }

    /**
     * Get average speed of ping (latency) from given ping output.
     *
     * @return average speed of ping (latency)
     */
    private int avgSpeed() {
        int start = src.lastIndexOf("=") + 2; // start of the number (2 chars after '=')
        if (start == 1) return -1; // before adding 2 was start -1 -> not found
        int end = start - 1;
        while (Character.isDigit(src.charAt(++end))) ;
        return Integer.parseInt(src.substring(start, end));
    }

    public boolean isOk() {
        return ok;
    }

    public int getAvgLatency() {
        return avgLatency;
    }

    public String getSrc() {
        return src;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Result merge(Result... results) {
        return this.mergeHelp(PingResult::getAvgLatency, avgSpeed -> new PingResult(null, avgSpeed, true), results);
    }

    /**
     * Static version of the {@link #merge(Result...)} method.
     *
     * @param results to be merged
     * @return the average result
     * @throws NotCompatibleException if the results cannot be merged
     */
    public static Result mergeResults(Result... results) throws NotCompatibleException {
        return results[0].merge(results);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean everythingOK() {
        return ok;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toHumanText() {
        return ConnectionQuality.get(avgLatency).toString() + ", avg latency: " + avgLatency + "ms";
    }
}
