package brambora;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import javafx.beans.property.FloatProperty;
import javafx.beans.property.ReadOnlyFloatProperty;
import javafx.beans.property.SimpleFloatProperty;

/**
 * @author brambora
 */
public class Speedtestdotnet {

    private final static int MAX_TIME_LIMIT_IN_SECONDS = 3;
    private final static int MIN_TIME_LIMIT_IN_SECONDS = 1;
    
    private final static int READ_BUFFER_SIZE = 1024;
    
    private final ByteBuffer readBuffer;
    
    //http://www.speedtest.net/speedtest-servers.php
    private final String serverUrl = "http://rychlost.giganet.cz/speedtest/";
    private final String uploadUrl = "upload.php";
    private final String latencyUrl = "latency.txt";
    
    private int fileToDownload = 4;
    private int uploadFileSizeInMegaBytes = 1;
    
    private static final LinkedList<String> filesToDownload;
    private final FloatProperty progress = new SimpleFloatProperty(-1);
    
    public Speedtestdotnet() {
        readBuffer = ByteBuffer.allocate(READ_BUFFER_SIZE);
    }
    
    public float testDownloadSpeed(int numberOfTests, int stripSize) throws IOException {
        return proceedMultipleTests(Speedtestdotnet::testDownloadSpeed, numberOfTests, stripSize);
    }

    public float testUploadSpeed(int numberOfTests, int stripSize) throws IOException {
        return proceedMultipleTests(Speedtestdotnet::testUploadSpeed, numberOfTests, stripSize);
    }

    public float testLatency(int numberOfTests, int stripSize) throws IOException {
        return proceedMultipleTests(Speedtestdotnet::testLatency, numberOfTests, stripSize);
    }

    private float proceedMultipleTests(ThrowableFunction<Speedtestdotnet, Float> test, int numberOfTests, int stripCount) throws IOException {
        if(numberOfTests - stripCount * 2 <= 0) {
            throw new RuntimeException("numberOfTests - (stripCount * 2) have to be greater than zero");
        }
        
        List<Float> results = new ArrayList<>(numberOfTests);

        for (int i = 1; i <= numberOfTests; i++) {
            results.add(test.apply(this));
            progress.set((float)i / numberOfTests);
        }

        removeWorstAndBestResults(results, stripCount);

        return (float) results.stream().mapToDouble(e -> e).sum();
    }

    public float testDownloadSpeed() throws MalformedURLException, IOException {
        URL url = new URL(serverUrl + filesToDownload.get(fileToDownload));
        ReadableByteChannel rbc = Channels.newChannel(url.openStream());

        long start, end;
        long fileSize = 0;

        start = System.currentTimeMillis();
        while (rbc.read(readBuffer) >= 0) {
            fileSize += readBuffer.position();
            readBuffer.clear();
        }
        end = System.currentTimeMillis();

        float timeInSeconds = (end - start) / 1000.0f;
        downloadOptimalization(timeInSeconds);
        
        return fileSize / timeInSeconds;
    }
    
    private void downloadOptimalization(float timeInSeconds) {
        if(timeInSeconds > MAX_TIME_LIMIT_IN_SECONDS) {
            if(--fileToDownload < 0) fileToDownload = 0;
        } 
        
        if(timeInSeconds < MIN_TIME_LIMIT_IN_SECONDS) {
            if(++fileToDownload > filesToDownload.size() - 1) fileToDownload = filesToDownload.size() - 1;
        } 
    }
    
    public float testUploadSpeed() throws MalformedURLException, IOException {
        URL url = new URL(serverUrl + uploadUrl);
        URLConnection con = url.openConnection();

        byte[] data = new byte[uploadFileSizeInMegaBytes * 1024 * 1024];
        con.addRequestProperty("Content-Type", "application/x-www-form-urlencoded");

        con.setDoOutput(true);

        long start, end;
        start = System.currentTimeMillis();
        try (OutputStream output = con.getOutputStream()) {
            output.write(data);
        }

        try (BufferedReader br = new BufferedReader(new InputStreamReader(con.getInputStream()))) {
            while (br.readLine() != null) {
            }
        }
        end = System.currentTimeMillis();

        float timeInSeconds = (end - start) / 1000.0f;
        int fileSize = data.length;
        uploadOptimalization(timeInSeconds);
        
        return fileSize / timeInSeconds;
    }

    private void uploadOptimalization(float timeInSeconds) {
        if(timeInSeconds > MAX_TIME_LIMIT_IN_SECONDS) {
            if(--uploadFileSizeInMegaBytes < 1) uploadFileSizeInMegaBytes = 1;
        } 
        
        if(timeInSeconds < MIN_TIME_LIMIT_IN_SECONDS) {
            uploadFileSizeInMegaBytes++;
        } 
    }
    
    public float testLatency() throws MalformedURLException, IOException {
        URL url = new URL(serverUrl + latencyUrl);

        long start, end;

        start = System.currentTimeMillis();
        try (BufferedReader br = new BufferedReader(new InputStreamReader(url.openStream()))) {
            while (br.readLine() != null) {
            }
        }
        end = System.currentTimeMillis();

        float timeInMillis = end - start;

        return timeInMillis;
    }

    private void removeWorstAndBestResults(List<Float> list, int stripCount) {
        Collections.sort(list);

        for (int i = 0; i < stripCount; i++) {
            list.remove(0);
            list.remove(list.size() - 1);
        }
    }

    public ReadOnlyFloatProperty getProgressProperty() {
        return progress;
    }
    
    static {
        filesToDownload = new LinkedList<>();
        filesToDownload.add("random350x350.jpg");
        filesToDownload.add("random500x500.jpg");
        filesToDownload.add("random750x750.jpg");
        filesToDownload.add("random1000x1000.jpg");
        filesToDownload.add("random1500x1500.jpg");
        filesToDownload.add("random2000x2000.jpg");
        filesToDownload.add("random2500x2500.jpg");
        filesToDownload.add("random3000x3000.jpg");
        filesToDownload.add("random3500x3500.jpg");
        filesToDownload.add("random4000x4000.jpg");
    }

}
